/* xpyre.c
 * xpyre effect logic and Allegro interface
 *
 * Copyright (C) 2002 Andy Goth <unununium@openverse.com>
 * For more information visit http://ioioio.net/devel/xpyre/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include <allegro.h>
#include <math.h>
#include <pthread.h>
#include "xpyre.h"
#include "font.h"

/* Prototypes */
static int set_graphics_mode(config_t*);
static void cleanup_graphics(config_t*);

/* Timer for maintaining a nearly-constant frame rate */
static volatile int timer = 0;
static void timer_proc(void) {timer++;}
END_OF_FUNCTION(timer_proc);

/* Fool Allegro into working */
int _mangled_main_address(int argc, char** argv) {}

/* Set the graphics mode given by s */
int set_graphics_mode(config_t* s)
{
	int mode = s->fullscreen ? GFX_AUTODETECT_FULLSCREEN :
			GFX_AUTODETECT_WINDOWED;
	int count = 0;
	int ret;

	/* Deallocate old double buffers */
	cleanup_graphics(s);

	/* Try up to three times to set the graphics mode */
	if (desktop_color_depth() != 0) set_color_depth(desktop_color_depth());
	do {
		ret = set_gfx_mode(mode, s->width, s->height, 0, 0);
		count++;
	} while (ret < 0 && count < 3);
	if (ret < 0) return 0;

	/* Create double buffers */
	s->bmp[0] = create_bitmap_ex(32, s->width, s->height);
	if (s->bmp[0] == NULL) return 0;
	s->bmp[1] = create_bitmap_ex(32, s->width, s->height);
	if (s->bmp[1] == NULL) {destroy_bitmap(s->bmp[0]); return 0;}
	clear_bitmap(s->bmp[0]); clear_bitmap(s->bmp[1]);

	return 1;
}

/* Free some bitmaps */
void cleanup_graphics(config_t* s)
{
	if (s->bmp[0] != NULL) destroy_bitmap(s->bmp[0]);
	if (s->bmp[1] != NULL) destroy_bitmap(s->bmp[1]);
}

/* Main loop for the xpyre effect; runs as its own thread */
void* main_loop(void* arg)
{
	track_t track = {-1, NULL};
	config_t mode = {320, 240, 0, 120, NULL, NULL};
	BITMAP* src, *dst;
	int flip = 1;
	int x, y;
	int show_title, title_width;
	float yscale;
	int fade;

	/* Fetch settings */
	load_config(&mode);

	/* Start Allegro */
	allegro_init();
	install_keyboard();
	install_timer();
	text_mode(-1);
	font = &my_font;

	/* Start the timer */
	LOCK_VARIABLE(timer);
	LOCK_FUNCTION(timer_proc);
	install_int_ex(timer_proc, BPS_TO_TIMER(mode.fps));

	/* Start the graphics stuff */
	set_window_title("xpyre");
	set_window_close_hook(quit);
	set_graphics_mode(&mode);
	yscale = 0.25 * mode.height / 240;
	fade = MAX(1, 8 * 240 / mode.height);

	/* Arrange for cleanup of all this stuff */
	pthread_cleanup_push((void(*)(void*))allegro_exit, NULL);
	pthread_cleanup_push((void(*)(void*))cleanup_graphics, &mode);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	/* Here it all is... */
	while (1) {
		/* Wait around for the timer to trigger */
		do {
			/* Handle the keyboard */
			if (!keypressed()) continue;
			switch (readkey() >> 8) {
			case KEY_ESC:
				mode.fullscreen = !mode.fullscreen;
				set_graphics_mode(&mode);
				yscale = 0.25 * mode.height / 240;
				fade = MAX(1, 8 * 240 / mode.height);
				break;
			case KEY_Z: previous_track(); break;
			case KEY_X: play_track(); break;
			case KEY_C: pause_track(); break;
			case KEY_V: stop_track(); break;
			case KEY_B: next_track(); break;
			}
		} while (timer == 0);

		/* Don't worry about making up for lost ticks */
		timer = 0;

		/* Set up the bitmap pointers */
		src = mode.bmp[flip];
		flip = 1 - flip;
		dst = mode.bmp[flip];

		/* Plot new pcm data on top of old image */
		for (x = 0; x < SCREEN_W; x++) {
			int c1, c2, v, s, o;

			v = abs(pcm_data[x * 512 / SCREEN_W] / 32);
			/* TODO: scale this up for higher resolution */
			y = -v * yscale + SCREEN_H - 1;

			/* Determine colors */
			v = MIN(v, 511);
			if (v < 256) {
				c1 = makecol32(v, 0, 0);
				c2 = makecol32(255, v, 0);
			} else {
				c1 = makecol32(255, v - 256, 0);
				c2 = makecol32(255, 255, v - 256);
			}

			/* Draw vertical line */
			vline(src, x, SCREEN_H - 1, y, c1);

			/* Draw smooth curve */
			if (x > 0) {
				s = (o + y) / 2;
				vline(src, x - 1, o, s, c2);
				vline(src, x, s, y, c2);
			}
			o = y;
		}

		/* Process image */
		for (y = 1; y < SCREEN_H - 1; y++) {
			unsigned* p1, *p2, *p3, *p2_beg, *p2_end, *q;
			int adj = rand() & 1 ? 0 : rand() & 1 ? 1 : -1;

			/* Initialize line pointers */
			if (adj >= 0) {
				/* Ripple to the right */
				p1 = (unsigned*)src->line[y - 1] + adj;
				p2 = (unsigned*)src->line[y] + adj;
				p3 = (unsigned*)src->line[y + 1] + adj;
				p2_beg = p2;
				p2_end = p2 + SCREEN_W - adj;
				q = (unsigned*)dst->line[y - 1];
			} else {
				/* Ripple to the left */
				p1 = (unsigned*)src->line[y - 1];
				p2 = (unsigned*)src->line[y];
				p3 = (unsigned*)src->line[y + 1];
				p2_beg = p2;
				p2_end = p2 + SCREEN_W + adj;
				q = (unsigned*)dst->line[y - 1] - adj;
			}

			/* Process scanline */
			while (p2 < p2_end) {
				int r, g, b;

				/* Blur */
				if (p2 == p2_beg) {
r = (((p1[0] & 0xff0000) >> 16) + ((p1[1] & 0xff0000) >> 16) +
((p2[0] & 0xff0000) >> 16) * 2 + ((p2[1] & 0xff0000) >> 16) +
((p3[0] & 0xff0000) >> 16) + ((p3[1] & 0xff0000) >> 16)) / 8;
g = (((p1[0] & 0x00ff00) >> 8) + ((p1[1] & 0x00ff00) >> 8) +
((p2[0] & 0x00ff00) >> 8) * 2 + ((p2[1] & 0x00ff00) >> 8) +
((p3[0] & 0x00ff00) >> 8) + ((p3[1] & 0x00ff00) >> 8)) / 8;
b = ((p1[0] & 0x0000ff) + (p1[1] & 0x0000ff) +
(p2[0] & 0x0000ff) * 2 + (p2[1] & 0x0000ff) +
(p3[0] & 0x0000ff) + (p3[1] & 0x0000ff)) / 8;
				} else if (p2 == p2_end - 1) {
r = (((p1[-1] & 0xff0000) >> 16) + ((p1[0] & 0xff0000) >> 16) +
((p2[-1] & 0xff0000) >> 16) * 2 + ((p2[0] & 0xff0000) >> 16) +
((p3[-1] & 0xff0000) >> 16) + ((p3[0] & 0xff0000) >> 16)) / 8;
g = (((p1[-1] & 0x00ff00) >> 8) + ((p1[0] & 0x00ff00) >> 8) +
((p2[-1] & 0x00ff00) >> 8) * 2 + ((p2[0] & 0x00ff00) >> 8) +
((p3[-1] & 0x00ff00) >> 8) + ((p3[0] & 0x00ff00) >> 8)) / 8;
b = ((p1[-1] & 0x0000ff) + (p1[0] & 0x0000ff) +
(p2[-1] & 0x0000ff) * 2 + (p2[0] & 0x0000ff) +
(p3[-1] & 0x0000ff) + (p3[0] & 0x0000ff)) / 8;
				} else {
r = (((p1[-1] & 0xff0000) >> 16) + ((p1[0] & 0xff0000) >> 16) * 2 +
((p1[1] & 0xff0000) >> 16) + ((p2[-1] & 0xff0000) >> 16) * 2 +
((p2[0] & 0xff0000) >> 16) * 4 + ((p2[1] & 0xff0000) >> 16) * 2 +
((p3[-1] & 0xff0000) >> 16) + ((p3[0] & 0xff0000) >> 16) * 2 +
((p3[1] & 0xff0000) >> 16)) / 16;
g = (((p1[-1] & 0x00ff00) >> 8) + ((p1[0] & 0x00ff00) >> 8) * 2 +
((p1[1] & 0x00ff00) >> 8) + ((p2[-1] & 0x00ff00) >> 8) * 2 +
((p2[0] & 0x00ff00) >> 8) * 4 + ((p2[1] & 0x00ff00) >> 8) * 2 +
((p3[-1] & 0x00ff00) >> 8) + ((p3[0] & 0x00ff00) >> 8) * 2 +
((p3[1] & 0x00ff00) >> 8)) / 16;
b = ((p1[-1] & 0x0000ff) + (p1[0] & 0x0000ff) * 2 +
(p1[1] & 0x0000ff) + (p2[-1] & 0x0000ff) * 2 +
(p2[0] & 0x0000ff) * 4 + (p2[1] & 0x0000ff) * 2 +
(p3[-1] & 0x0000ff) + (p3[0] & 0x0000ff) * 2 +
(p3[1] & 0x0000ff)) / 16;
				}

				/* Fade */
				/* TODO: make fade slower for high resolution */
				if (b > 0) b = MAX(0, b - fade);
				else if (g > 0) g = MAX(0, g - fade);
				else if (r > 0) r = MAX(0, r - fade);
				*q = b | (g << 8) | (r << 16);

				/* Blit the pixel! */
				p1++; p2++; p3++; q++;
			}
		}

		/* Check for song change */
		if (update_track(&track)) {
			show_title = 30;
			title_width = text_length(font, track.title) + 5;
		}

		/* Display the title immediately after a song change */
		if (show_title != 0) {
			if (show_title > 0) {
				x = 5;
				show_title--;
				if (show_title == 0) show_title = -1;
			} else {
				x = show_title;
				show_title -= 2;
				if (show_title <= -title_width) show_title = 0;
			}
			textout(dst, font, track.title, x + 1, 17, 0x000000);
			textout(dst, font, track.title, x, 16, 0xffffff);
		}

		/* The Big Blit! */
		blit(dst, screen, 1, 0, 1, 1, SCREEN_W - 2, SCREEN_H - 2);
	}

	pthread_cleanup_pop(0);
	pthread_cleanup_pop(0);
}
